/*
Copyright  2004-2006 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   Author scam+grey@elandsys.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <syslog.h>

#include <pthread.h>

#include "libmilter/mfapi.h"

#ifndef LINUX
#include <sys/queue.h>
#else
#include "queue.h"
#endif

#include "lock.h"
#include "util.h"

#define PIDFILE "/var/spool/scam/scam-grey.pid"
#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))

#define VERSION "1.4"

/* POPAUTH db */
#ifdef POPAUTH
# define POPAUTHDB	"/etc/mail/popip.db"
#endif
#define OSSOCK "/var/spool/scam/p0f.sock"
#define QUERY_MAGIC		0x0defaced

typedef unsigned char		_u8;
typedef unsigned short		_u16;
typedef unsigned int		_u32;

typedef unsigned long long	_u64;

typedef signed char		_s8;
typedef signed short		_s16;
typedef signed int		_s32;

typedef signed long long	_s64;

#define QTYPE_FINGERPRINT       1

struct p0f_query {
  _u32 magic;			/* must be set to QUERY_MAGIC */
#ifndef P0F206
  _u8  type;			/* QTYPE_* */
#endif
  _u32 id;			/* Unique query ID */
  _u32 src_ad,dst_ad;		/* src address, local dst addr */
  _u16 src_port,dst_port;	/* src and dst ports */
};

#define RESP_OK		0	/* Response OK */
#define RESP_BADQUERY	1	/* Query malformed */
#define RESP_NOMATCH	2	/* No match for src-dst data */

struct p0f_response {
  _u32 magic;			/* QUERY_MAGIC */
  _u32 id;			/* Query ID (copied from p0f_query) */
  _u8  type;			/* RESP_* */
  _u8  genre[20];		/* OS genre (empty if no match) */
  _u8  detail[40];		/* OS version (empty if no match) */
  _s8  dist;			/* Distance (-1 if unknown ) */
  _u8  link[30];		/* Link type (empty if unknown) */
  _u8  tos[30];			/* Traffic type (empty if unknown) */
  _u8  fw,nat;			/* firewall and NAT flags flags */
  _u8  real;			/* A real operating system? */
  _s16 score;			/* Masquerade score (or NO_SCORE) */
  _u16 mflags;			/* Masquerade flags (D_*) */
  _s32 uptime;			/* Uptime in hours (-1 = unknown) */
};

struct greyent {
	char	*helo;
	struct in_addr addr;
	struct timeval tv;
	unsigned short int allow;
	TAILQ_ENTRY(greyent)	greyentries;
};

pthread_rwlock_t conn_lock;
pthread_rwlock_t grey_lock;

TAILQ_HEAD(greylist, greyent);
struct greylist greyhead;

static volatile unsigned int connections = 0;
static int testmode = 0;
static int rdns = 0;
char *daemonmta = NULL;
char *macro = NULL;
char *gmacro = NULL;
char *contact = NULL;

struct mlfiPriv {
	char	*connectfrom;
	struct in_addr	hostaddr;
	char  *if_name;
	int	clientport;
	struct in_addr serveraddr;
	char	*helo;
	char	*queueid;
	int	msa;
	int	smtpauth;
	int	distance;
	char	*os;
	int	nordns;
	int	greylist;
	int	asn;
} mlfi_priv;


static sfsistat mlfi_cleanup(SMFICTX *);

static sfsistat
upgreylist(SMFICTX *ctx)
{
	struct mlfiPriv *priv= MLFIPRIV;
	sfsistat rstat = SMFIS_ACCEPT;

	if (((priv->msa == 0) && (priv->smtpauth == 0) && (priv->distance >= greylistdistance) && (priv->os != NULL) && ((strncasecmp("windows", priv->os, 7) == 0) || (priv->nordns == 1))) ||  (priv->greylist == 1))
	{
		struct greyent *gent;
		struct greyent *gentnext;
		struct greyent *gentadd;
		int greyadd = 0;
		char scanid[256];
		char smreply[256];
		struct timeval gnow;
#ifndef LINUX
		struct timezone tz;

		gettimeofday(&gnow, &tz);
#else
		gettimeofday(&gnow, NULL);
#endif

		WRLOCK(grey_lock);
		for (gent = TAILQ_FIRST(&greyhead); gent; gent = gentnext)
		{
			gentnext = TAILQ_NEXT(gent, greyentries);

			if ((gent->addr.s_addr == priv->hostaddr.s_addr ) && (strcasecmp(gent->helo, priv->helo) == 0))
			{
				greyadd = 1;
				if (gnow.tv_sec - gent->tv.tv_sec < greylisttime)
				{
					syslog (LOG_INFO, "%s: greylisted %s", priv->queueid, priv->connectfrom);

					if (contact == NULL)
					{
						snprintf( smreply, 255, "Greylisted by scam-grey http://www.elandsys.com/scam/ - Please retry after %d minutes", (greylisttime/60) );
					} else {
						snprintf( smreply, 255, "Greylisted by scam-grey http://www.elandsys.com/scam/ - Please retry after %d minutes - %s", (greylisttime/60), contact);
					}
					smfi_setreply( ctx, "450", "4.7.0", smreply);
					if (testmode == 0)
					{
						rstat = SMFIS_TEMPFAIL;
					} else {
#ifdef SM813
					smfi_insheader(ctx, 1, "X-Scam-grey", "positive");
#else
					smfi_addheader(ctx, "X-Scam-grey", "positive");
#endif
					}
				} else {
					gent->allow = 1;
					snprintf( scanid, sizeof(scanid), "delay %d seconds", greylisttime);
#ifdef SM813
					smfi_insheader(ctx, 1, "X-Scam-grey", scanid);
#else
					smfi_addheader(ctx, "X-Scam-grey", scanid);
#endif
				}

				break;
			}

			if ((gnow.tv_sec - gent->tv.tv_sec > greylistexpiry) && (gent->allow == 0))
			{
				syslog (LOG_INFO, "%s: removed %s from greylist", priv->queueid, inet_ntoa(gent->addr));
				TAILQ_REMOVE(&greyhead, gent, greyentries);
				free(gent->helo);
				free(gent);
				continue;
			}

		}

		UNLOCK(grey_lock);

		if (greyadd == 0)
		{
			gentadd = (struct greyent *)malloc(sizeof(struct greyent));
			if ((gentadd->helo = strdup(priv->helo)) == NULL)
			{
				syslog (LOG_INFO, "%s: greylist cannot alloc", priv->queueid);
				rstat = SMFIS_TEMPFAIL;
			} else {
				gentadd ->addr.s_addr = priv->hostaddr.s_addr;
				gentadd->tv = gnow;
				gentadd->allow = 0;
				WRLOCK(grey_lock);
				TAILQ_INSERT_TAIL(&greyhead, gentadd, greyentries);
				UNLOCK(grey_lock);
				syslog (LOG_INFO, "%s: added %s to greylist", priv->queueid, priv->connectfrom);
				if (contact == NULL)
				{
					snprintf( smreply, 255, "Greylisted by scam-grey http://www.elandsys.com/scam/ - Please retry after %d minutes", (greylisttime/60) );
				} else {
					snprintf( smreply, 255, "Greylisted by scam-grey http://www.elandsys.com/scam/ - Please retry after %d minutes - %s", (greylisttime/60), contact);
				}
				smfi_setreply( ctx, "450", "4.7.0", smreply);
				if (testmode == 0)
				{
					rstat = SMFIS_TEMPFAIL;
				} else {
#ifdef SM813
				smfi_insheader(ctx, 1, "X-Scam-grey", "positive");
#else
				smfi_addheader(ctx, "X-Scam-grey", "positive");
#endif
				}
			}
		}
	}
	return (rstat);
}

static sfsistat
mlfi_connect(SMFICTX *ctx, char *hostname,  _SOCK_ADDR *hostaddr)
{
	struct mlfiPriv *priv= MLFIPRIV;
	struct sockaddr_in  *remoteaddr;
	char *daemonname;
	char *ifaddr = NULL;
	char  *clientresolve;

	priv = malloc(sizeof *priv);
	if (priv == NULL) {
		/* can't accept this message right now */
		syslog (LOG_ERR, "cannot allocate priv");
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	/* set the private data */
	if (smfi_setpriv(ctx, priv) != MI_SUCCESS)
	{
		smfi_setpriv(ctx, priv);
		free(priv);

		syslog (LOG_ERR, "cannot setpriv");
		(void) mlfi_cleanup(ctx);
        return SMFIS_TEMPFAIL;
	}

	priv->queueid = NULL;

	WRLOCK(conn_lock);
	connections++;
	UNLOCK(conn_lock);

	remoteaddr = (struct sockaddr_in *) hostaddr;

	if ((remoteaddr != NULL) && (remoteaddr->sin_family == AF_INET))
	{
		/* FIX ME inet_addr */
		if (remoteaddr->sin_addr.s_addr == inet_addr("127.0.0.1"))
			return SMFIS_ACCEPT;

		priv->hostaddr.s_addr = remoteaddr->sin_addr.s_addr;
		priv->clientport = ntohs(remoteaddr->sin_port);
	}

	if ((hostname)  && ((priv->connectfrom= strdup (hostname)) == NULL))
	{
		syslog (LOG_ERR, "cannot allocate connect");
        return SMFIS_TEMPFAIL;
    }

	daemonname = smfi_getsymval(ctx, "{daemon_name}");
	if (daemonname != NULL)
	{
		if (strcasecmp(daemonname, "MSA") == 0)
		{
				priv->msa = 1;
		} else {

			if (daemonmta != NULL)
			{
				if (strcasecmp(daemonname, daemonmta) == 0)
				{
					priv->msa = 1;
				}
			}
#if POPAUTH
			if (priv->msa == 0)
				priv->msa = checkpopauth(POPAUTHDB, hostaddr);
#endif
		}

	}


	if (priv->msa == 1)
	{
		return SMFIS_ACCEPT;
	}

	ifaddr = smfi_getsymval(ctx, "{if_addr}");
	if (ifaddr != NULL)
	{
		priv->serveraddr.s_addr = inet_addr(ifaddr);
	}

	if (rdns == 1)
	{
		clientresolve = smfi_getsymval(ctx, "{client_resolve}");
		if (clientresolve != NULL) {
			if (strcasecmp(clientresolve, "FAIL") == 0) {
				priv->nordns = 1;
			}
		}
	}

	return SMFIS_CONTINUE;
}

sfsistat
mlfi_helo(SMFICTX *ctx, char *helohost)
{
	struct mlfiPriv *priv = MLFIPRIV;
	int chkre = 0;

	if (check_ip(priv->hostaddr.s_addr) == 1)
	{
		return SMFIS_ACCEPT;
	}

	if(helohost == NULL)
	{
		syslog (LOG_ERR, "missing helo");
#ifdef NOHELOHOST
		if ((priv->helo = strdup (".")) == NULL)
		{
			syslog (LOG_ERR, "cannot allocate helo");
			return SMFIS_TEMPFAIL;
		}
#else
		return SMFIS_TEMPFAIL;
#endif
	} else if ((priv->helo = strdup (helohost)) == NULL)
	{
		syslog (LOG_ERR, "cannot allocate helo");
        return SMFIS_TEMPFAIL;
    }

	if (priv->connectfrom)
	{
		if (dothost(priv->connectfrom) < 3)
			return SMFIS_ACCEPT;
	}

	if (priv->nordns == 0)
	{
		chkre = check_re(priv->connectfrom, priv->helo);

		if (chkre  == 1)
			return SMFIS_ACCEPT;
		else if (chkre  == 2)
			priv->greylist = 1;
	}

	if (priv->msa == 0)
	{
		priv->asn = qrydnsasn( priv->hostaddr.s_addr);
		if (priv->asn > 0)
		{
			chkre = check_asn(priv->asn);

			if (chkre  == 1)
				return SMFIS_ACCEPT;
			else if (chkre  == 2)
				priv->greylist = 1;
		}
	}

	if (( priv->msa == 0) && (priv->greylist == 0))
	{
		struct p0f_query q;
		struct p0f_response r;
		int sockfd = 0;
		int rc;

		if (priv->serveraddr.s_addr > 0)
		{
			q.magic    = QUERY_MAGIC;
			q.id       = connections;
#ifndef P0F206
			q.type      = QTYPE_FINGERPRINT;
#endif
			q.src_ad   = priv->hostaddr.s_addr;
			q.dst_ad   =  priv->serveraddr.s_addr;
			q.src_port = priv->clientport;
			q.dst_port = 25;

			sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
		} else
			syslog (LOG_ERR, "cannot get interface address");

		if (sockfd > 0)
		{
			rc = unixconn( sockfd, OSSOCK);
			if (rc == 0)
			{
				if (write(sockfd,&q,sizeof(q)) != sizeof(q))
					syslog(LOG_ERR, "p0f write error");
				else {
					if (read(sockfd,&r,sizeof(r)) == sizeof(r))
					{
						switch (r.magic)
						{
							case QUERY_MAGIC:
								priv->distance = r.dist;
								if ((priv->os = strdup(r.genre)) == NULL)
								{
									syslog (LOG_ERR, "cannot allocate os identity");
								}
								break;

							case RESP_BADQUERY:
								syslog( LOG_INFO, "bad query");
								break;

							case RESP_NOMATCH:
								syslog( LOG_INFO, "no match");
								break;

							default:
								syslog( LOG_INFO, "no response");
						}
					 } else
						syslog(LOG_ERR, "p0f cannot read socket");
				 }
			}
			shutdown(sockfd, 2);
			close(sockfd);
		}
	}

	return SMFIS_CONTINUE;
}


static sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	struct mlfiPriv *priv= MLFIPRIV;
	char *authtype;
	char *queueid;
	char *macroval;
	char *gmacroval;

	authtype = smfi_getsymval(ctx, "{auth_type}");

	if (authtype != NULL)
	{
		priv->smtpauth = 1;
	}

	queueid= smfi_getsymval(ctx, "i");
	if (queueid != NULL)
	{
		if (priv->queueid != NULL)
			free(priv->queueid);

		if ((priv->queueid = strdup (queueid)) == NULL)
		{
			syslog (LOG_ERR, "cannot allocate queueid");
			(void) mlfi_cleanup(ctx);
			return SMFIS_TEMPFAIL;
		}
	}

	if ((priv->smtpauth == 0) && (priv->msa == 0))
	{
		if (macro != NULL)
		{
			macroval = smfi_getsymval(ctx, macro);
			if (macroval != NULL)
			{
				priv->msa = 1;
				syslog (LOG_INFO, "%s: %s whitelist", priv->queueid, macro);
			}
		}

		if (gmacro != NULL)
		{
			gmacroval = smfi_getsymval(ctx, gmacro);
			if (gmacroval != NULL)
			{
				priv->greylist = 1;
				syslog (LOG_INFO, "%s: %s greylist", priv->queueid, gmacro);
			}
		}
	}

#ifdef CHKMF
	return upgreylist(ctx);
#else
	return SMFIS_CONTINUE;
#endif
}

#ifdef CHKRCPT
static sfsistat
mlfi_envrcpt(SMFICTX *ctx, char **argv)
{
	return upgreylist(ctx);
}
#endif

static sfsistat
mlfi_eoh(SMFICTX *ctx)
{
#ifdef CHKRCPT
	return SMFIS_CONTINUE;
#endif
#ifdef CHKEOH
	return upgreylist(ctx);
#else
	return SMFIS_CONTINUE;
#endif
}

static sfsistat
mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv= MLFIPRIV;
	char asntag[10];
	if (priv->asn > 0)
	{
		snprintf( asntag, 9, "ASN%d", priv->asn);
#ifdef SM813
		smfi_insheader(ctx, 1, "X-Scam-ASN", asntag);
#else
		smfi_addheader(ctx, "X-Scam-ASN", asntag);
#endif
	}
#ifdef CHKRCPT
	return SMFIS_CONTINUE;
#endif
#ifndef CHKEOH
	return upgreylist(ctx);
#else
	return SMFIS_CONTINUE;
#endif
	return upgreylist(ctx);
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{

	mlfi_cleanup(ctx);

	return SMFIS_CONTINUE;
}

static sfsistat
mlfi_cleanup(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	if (priv == NULL)
		return SMFIS_CONTINUE;

	if (priv->queueid)
	{
		free(priv->queueid);
		priv->queueid = NULL;
	}

	return SMFIS_CONTINUE;

}

static sfsistat
mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	mlfi_cleanup(ctx);

	if (priv->connectfrom)
	{
		free(priv->connectfrom);
		priv->connectfrom = NULL;
	}

	if (priv->helo)
	{
		free(priv->helo);
		priv->connectfrom = NULL;
	}

	if (priv->if_name)
	{
		free(priv->if_name);
		priv->if_name = NULL;
	}

	if (priv->os)
	{
		free(priv->os);
		priv->os = NULL;
	}

	if (priv)
	{
		free(priv);
	}
	smfi_setpriv(ctx, NULL);

	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"scam-grey",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS,	/* flags */
	mlfi_connect,		/* connection info filter */
	mlfi_helo,		/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
#ifdef CHKRCPT
	mlfi_envrcpt,	/* envelope recipient filter */
#else
	NULL,		/* envelope recipient filter */
#endif
	NULL,				/* header filter */
	mlfi_eoh,	/* end of header */
	NULL,			/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close,	/* connection cleanup */
};

static void
usage()
{
	fprintf(stderr, "scam-grey version %s\n", VERSION);
    fprintf(stderr, "usage: scam-grey -p protocol:address [-T timeout] [-m daemonname] [-M macro] [-C message] [-R] [-t] [-D]\n");
	fprintf(stderr, "          -C append message to SMTP reply\n");
	fprintf(stderr, "          -D run as a daemon\n");
	fprintf(stderr, "          -G specify a macro to greylist\n");
	fprintf(stderr, "          -m specify a daemonname to whitelist\n");
	fprintf(stderr, "          -M specify a macro to whitelist\n");
	fprintf(stderr, "          -R greylist host with no rDNS\n");
	fprintf(stderr, "          -t run in test mode\n");
	fprintf(stderr, "          -u run as specified user\n");
}

static void
catch_signal( int signo )
{
	switch (signo)
	{
		case SIGUSR1:
		   syslog (LOG_NOTICE, "Caught SIGUSR1, exiting...");
			smfi_stop();
			sleep(10);
			exit (0) ;
	}

}

int
write_pid()
{
	char buf[20];
	int fd;
	long pid;

	if ((fd = open (PIDFILE, O_CREAT | O_TRUNC | O_WRONLY, 0600)) == -1)
	{
		syslog (LOG_ERR, "cannot open pidfile ");
	} else {
		pid = getpid ();
		snprintf (buf, sizeof (buf), "%ld", (long) pid);
		if (write (fd, buf, strlen (buf)) != strlen (buf)) {
			syslog (LOG_INFO, "cannot write to pidfile ");
		}
		close(fd);
	}
	return pid;
}

void
daemonize()
{

	signal (SIGUSR1, catch_signal);

	(void)close(0);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(1);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(2);
	(void)open("/dev/null", O_RDONLY, 0);

	switch (fork ())
	{
		case 0:
			syslog (LOG_INFO, "scam-grey running ");
			break;

		case -1:
			perror ("cannot fork\n");
			exit (3);
			break;

		default:
			exit(0);
	}

	if (setsid() < 0)
		exit(1);

	write_pid();
	chdir("/");

}

int
main(int argc, char *argv[])
{
	int c;
	int ret;
	const char *args = "p:T:h:u:g:m:M:C:G:RtD";
	extern char *optarg;
	int daemonmode = 0;
	struct passwd *passwd = NULL;
	struct group *grp;
	char *user = NULL;
	char *group = NULL;
	uid_t uid;
	gid_t gid;

	umask(077);

	/* already a daemon */
	if(getppid()==1)
		return 0;

	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != (char)EOF)
	{
		switch (c)
		{
		  case 'p':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal connection: %s\n", optarg);
				exit(EX_USAGE);
			}
			if (smfi_setconn(optarg) == MI_FAILURE)
			{
				(void) fprintf(stderr, "smfi_setconn failed");
				exit(EX_SOFTWARE);
			}

			/*
				** If we're using a local socket, make sure it doesn't
				** already exist.
			*/
            if(strncmp(optarg, "unix:", 5) == 0)
                unlink(optarg + 5);
            else if(strncmp(optarg, "local:", 6) == 0)
                unlink(optarg + 6);

			break;

		  case 'T':
            if (optarg == NULL || *optarg == '\0')
            {
                (void) fprintf(stderr, "Illegal timeout: %s\n", optarg);
                exit(EX_USAGE);
            }
            if(smfi_settimeout(atoi(optarg)) == MI_FAILURE)
            {
                (void) fputs("smfi_settimeout failed", stderr);
                exit(EX_SOFTWARE);
            }
            break;

			case 'u':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid username: %s\n", optarg);
					exit(EX_USAGE);
				}
				user = strdup(optarg);
				if ((passwd = getpwnam(user)) != NULL)
				{
					uid = passwd->pw_uid;
					gid = passwd->pw_gid;
				}
				break;

			case 'g':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid group name: %s\n", optarg);
					exit(EX_USAGE);
				}
				group = strdup(optarg);
				if ((grp = getgrnam(group)) != NULL)
				{
					gid = grp->gr_gid;
				}
				break;

		  case 'd':
			if (optarg == NULL || *optarg == '\0')
            {
                (void) fprintf(stderr, "Illegal debug value: %s\n", optarg);
                exit(EX_USAGE);
            }
			if (smfi_setdbg(atoi(optarg)) == MI_FAILURE)
				fprintf(stderr, "smfi_setdbg enabled\n");
			break;

		  case 'D':
			  daemonmode = 1;
			  break;

		  case 'm':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid daemonname: %s\n", optarg);
					exit(EX_USAGE);
				}
				daemonmta = strdup(optarg);
				break;

		case 'M':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid whitelist macro: %s\n", optarg);
					exit(EX_USAGE);
				}
				macro = strdup(optarg);
				break;

		case 'G':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid greylist macro: %s\n", optarg);
					exit(EX_USAGE);
				}
				gmacro = strdup(optarg);
				break;

		case 'C':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid contact: %s\n", optarg);
					exit(EX_USAGE);
				}
				if (strlen(optarg) > 100)
				{
					(void) fprintf(stderr, "Contact should be less than 100 characters: %s\n", optarg);
					exit(EX_USAGE);
				}
				contact = strdup(optarg);
				break;

		  case 'R':
			  rdns = 1;
			  break;

		  case 't':
			  testmode = 1;
			  break;

		  case 'h':
		  default:
			usage();
			exit(0);

		}
	}

	(void) closelog();
	openlog("scam-grey", LOG_PID | LOG_NDELAY, LOG_MAIL);

	setgid(gid);
	setuid(uid);

	if (!getuid() || !geteuid())
	{
		syslog (LOG_ERR, "scam-grey cannot run as root");
		exit(0);
	}

	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "smfi_register failed\n");
		syslog (LOG_ERR, "smfi_register failed");
		exit(EX_UNAVAILABLE);
	}

	TAILQ_INIT(&greyhead);

	if ((ret = pthread_rwlock_init(&conn_lock, NULL)) != 0)
	{
		syslog(LOG_ERR,
		    "pthread_rwlock_init failed: %s", strerror(ret));
		exit(EX_OSERR);
	}

	if ((ret = pthread_rwlock_init(&grey_lock, NULL)) != 0)
	{
		syslog(LOG_ERR,  "pthread_rwlock_init failed: %s", strerror(ret));
		exit(EX_OSERR);
	}

	gwhite_init();
	whitere_init();
	greyhostre_init();
	gwhite_readconf();

	if (daemonmode == 1)
	{
		daemonize();
	}

	ret = smfi_main();
	unlink(PIDFILE);
	syslog (LOG_INFO, "Exit");
	return(ret);
}

